#!/usr/bin/perl -w

# Набор тестов для модуля Jevix::MakeupUtf

use strict;
use warnings;
use Jevix::MakeupUtf;
use Data::Dumper;

# Русская локаль=================================================
use POSIX;
POSIX::setlocale (&POSIX::LC_CTYPE, "ru");
use locale;
# ===============================================================

# Список тестов вопрос-ответ
my @tests = (
		{ 
			q=>'Проверям тире, лишние пробелы, многоточие ... Слово   - дело',
			a=>'Проверям тире, лишние пробелы, многоточие&hellip; Слово&nbsp;&mdash; дело'
		},

		{
			q=>'Проверяем "кавычки внешние". После этого он сказал: "Не моё это "дело""',
			a=>'Проверяем &laquo;кавычки внешние&raquo;. После этого он сказал: &laquo;Не моё это &ldquo;дело&rdquo;&raquo;'
		},

		{
			q=>'"кавычки с <b>"тегами"</b>"',
			a=>'&laquo;кавычки с <b>&ldquo;тегами&rdquo;</b>&raquo;'
		},

		{
			q=>'Составное слово как-то так',
			a=>'Составное слово <nobr>как-то</nobr> так'
		},

		{
			addConf=>{'paragraphs'=>1},
			q=>"параграф1\n\nпараграф2\nновая строка",
			a=>"<p>параграф1</p>\n\n<p>параграф2<br />\nновая строка</p>"
		},

		{
			addConf=>{'paragraphs'=>0},
			q=>"-- До свиданья, внучек, до свиданья! – закивала бабка, – Вежливый какой, сейчас таких нет. Вот, например, Петрович – помню, помню, молодой был, галантный, воспитанный…\n-- До свиданья, внучек, до свиданья! – закивала бабка, – Вежливый какой, сейчас таких нет. Вот, например, Петрович – помню, помню, молодой был, галантный, воспитанный…\n-- До свиданья, внучек, до свиданья! – закивала бабка, – Вежливый какой, сейчас таких нет. Вот, например, Петрович – помню, помню, молодой был, галантный, воспитанный…",
			a=>"&mdash;&nbsp;До свиданья, внучек, до&nbsp;свиданья! – закивала бабка, – Вежливый какой, сейчас таких нет. Вот, например, Петрович – помню, помню, молодой был, галантный, воспитанный&hellip;<br />\n&mdash;&nbsp;До свиданья, внучек, до&nbsp;свиданья! – закивала бабка, – Вежливый какой, сейчас таких нет. Вот, например, Петрович – помню, помню, молодой был, галантный, воспитанный&hellip;<br />\n&mdash;&nbsp;До свиданья, внучек, до&nbsp;свиданья! – закивала бабка, – Вежливый какой, сейчас таких нет. Вот, например, Петрович – помню, помню, молодой был, галантный, воспитанный&hellip;"
		},

		{
			q=>'Тест <a href="http://jevix.ru" title=hello onload="alert(1)">допустимых</a> тегов и <b     cLaSs= my style="wow">аттрибутов</b>',
			a=>'Тест <a href="http://jevix.ru" title=hello>допустимых</a> тегов и <b style="wow">аттрибутов</b>'
		},

		{
			q=>'Здесь должна быть ссылка: http://jevix.ru, <a href="http://jevix.ru">а здесь - нет http://jevix.ru</a>',
			a=>'Здесь должна быть ссылка: <a href="http://jevix.ru" target="_blank">http://jevix.ru</a>, <a href="http://jevix.ru">а&nbsp;здесь&nbsp;&mdash; нет http://jevix.ru</a>'
		},

		{
			addConf=>{'tagNamesToLower'=>1},
			q=>'Приводим всё к <B>нижнему</b> <a HREF="http://jevix.ru">регистру</a>',
			a=>'Приводим всё к <b>нижнему</b> <a href="http://jevix.ru">регистру</a>'
		},

		{
			addConf=>{'tagsAllow'=>'a,br,b'},
			q=>'XSS test <script>alert(1)</script>. <a href="http://jevix.ru" onmouseOVER="alert(1)">hey!</a>',
			a=>'XSS test. <a href="http://jevix.ru">hey!</a>'
		},

		{
			addConf=>{'tagQuoteValues'=>1},
			q=>'<h1>this is test</h1><a href=jevix.ru title = link class="my">try jevix</a>',
			a=>'<h1>this is test</h1><a href="jevix.ru" title = "link" class="my">try jevix</a>'
		},
		
		{
			addConf=>{'tagCloseOpen'=>1},
			q=>'<p><h1>this is test',
			a=>'<p><h1>this is test</h1></p>'
		},

	      );

# Настройка при которой проводятся тесты
my $conf = {
               isHTML=>1,							# Работать в режиме гипертекста (в режиме простого текста работает быстрее)
               vanish=>0,							# Преобразовать исходный текст в plain-text (игонорирует все прочие настройки)               
               lineBreaks=>1,							# Расставлять переносы строк <br />
               paragraphs=>0,							# Размечать параграфы <p>
               dashes=>1,							# Тире
               dots=>1,								# Многоточия
               edgeSpaces=>1,							# Убирать пробельные символы в начале и конце строки
               tagSpaces=>1,							# Убирать пробелы между тегами (</td>  <td>)
               multiSpaces=>1,							# Превращать множественные пробелы в одинарные
               redundantSpaces=>1,						# Убирать пробелы там, где их не должно быть
               compositeWords=>1,						# Заключать составные слова в тег <nobr>
               compositeWordsLength=>10,					# Максимальная длина составного слова, заключаемого в тег <nobr>
               nbsp=>1,								# Расставлять неразрывные пробелы
               quotes=>1,							# Верстать кавычки
               qaType=>0,							# Тип внешних кавычек (см. настройки отладочного интерфейса на http://jevix.ru/)
               qbType=>1,							# Тип вложенных кавычек
               misc=>1,								# Всякое разное (&copy, дроби и прочее)
               codeMode=>2,							# Способ кодировки спец. символов (0: ANSI <...>, 1: HTML-код <&#133;>, 2: HTML-сужности <&hellip;>)
               tagsDenyAll=>0,							# По умолчанию отвергать все теги
               tagsDeny=>'',							# Список запрещённых тегов
               tagsAllow=>'|A:href:title,br,B:STYLE',				# Список разрешённых тегов (исключает их, когда устанавлен запрет всех)
               tagCloseSingle=>0,						# Закрывать одинарные теги, когда они не закрыты
	       tagCloseOpen=>0,							# Закрывать открытые теги в конце документа
               tagNamesToLower=>0,						# Приводить имена тегов к нижнему регистру
               tagNamesToUpper=>0,						# Приводить имена тегов к верхнему регистру
               tagAttributesToLower=>0,						# Приводить имена атрибутов тегов к нижнему регистру
               tagAttributesToUpper=>0,						# Приводить имена атрибутов тегов к верхнему регистру
               tagQuoteValues=>0,						# Помещать в кавычки значения атрибутов тегов
               tagUnQuoteValues=>0,						# Убирать кавычки вокруг значений атрибутов тегов
               links=>1,							# Автоматически заключать ссылки в тег <a>
               linksAttributes=>{target=>'_blank'},				# Хеш аттрибутов для всех новых ссылок
               simpleXSS=>1,							# Удаление возможных XSS-атак в коде документа
               checkHTML=>0,							# Проверять целостность HTML
               logErrors=>0							# Вести журнал ошибок
};

my $text;
my $result;
my $testsCount = 0;
my $errCount = 0;

my $jevix = new Jevix::MakeupUtf;

print "\n\nTesting Jevix Class...\n\n";

# Тест парсера строки настройки фильтра HTML
$testsCount++;
$jevix->setConf($conf);

my $pTest = $jevix->getConf();
$pTest = $pTest->{tagsAllow};

if(!($pTest->{br} && $pTest->{b} && $pTest->{a} && $pTest->{a}->{validAttributes}->{href} && $pTest->{a}->{validAttributes}->{title})) {
	print "Allowed tags string parser failure\n\n";
	$errCount++;
}

# Прогоняем текстовые тесты
foreach my $test (@tests) {
	$testsCount++;
	$text = $test->{q};

	# Изменение конфигурации
	if($test->{addConf}) {
		while (my ($k,$v) = each(%{$test->{addConf}})) {
			$conf->{$k} = $v
		}
		
		$jevix->setConf($conf);
	}

	$result = $jevix->process(\$text);

	if($test->{a} ne $result->{text}) {
		print 'Test ' . $testsCount . ' failed: [' . $text . "]\n";
		print 'Expected: [' . $test->{a} . "]\n";
		print 'We have:  [' . $result->{text} . "]\n\n";
		$errCount++;
	}
}

# Статистика
print "Tests Perfomed: " . $testsCount . "\n";
print "Tests Failed: " . $errCount . "\n";
print "Well Done!\n\n" if(!$errCount);